from .parameters import require_string, require_bytes
from .mixins.mixin_base import MixinBaseClass


class RedisEntityMetaclass(type):
    """
    This metaclass makes sure that all MixinClasses are placed before other classes in
    the mro of the new class. This will then enable the functionality of overriding
    classmethod in the Mixin classes.
    """

    def __new__(mcs, name, bases, namespace):
        mixin_bases = tuple(
            mixin_class
            for mixin_class in bases
            if issubclass(mixin_class, MixinBaseClass)
        )
        other_bases = tuple(
            other_class for other_class in bases if other_class not in mixin_bases
        )
        bases = (*mixin_bases, *other_bases)
        return super().__new__(mcs, name, bases, namespace)


class BaseRedisEntity(metaclass=RedisEntityMetaclass):
    RedisClient = None
    Prefix = ""
    Expire = None

    @classmethod
    def _hash_identifier(cls, identifier) -> str:
        return require_string(identifier)

    @classmethod
    def build_identifier(cls, identifier):
        return f"{cls.Prefix}:{require_string(cls._hash_identifier(identifier))}"

    @classmethod
    def encrypt_data(cls, key, value) -> bytes:
        return require_bytes(value)

    @classmethod
    def decrypt_data(cls, key, value) -> bytes:
        return require_bytes(value)

    @classmethod
    def add_expiration(cls, identifier):
        if isinstance(cls.Expire, int) and cls.Expire >= 0:
            cls.RedisClient.expire(identifier, cls.Expire)



from .base_entity import BaseRedisEntity


class RedisSetEntity(BaseRedisEntity):
    @classmethod
    def add(cls, identifier, *values):
        identifier = cls.build_identifier(identifier)
        values = [cls.encrypt_data(identifier, value) for value in values]
        cls.RedisClient.sadd(identifier, *values)
        cls.add_expiration(identifier)

    @classmethod
    def delete(cls, identifier, *values):
        identifier = cls.build_identifier(identifier)
        values = [cls.encrypt_data(identifier, value) for value in values]
        cls.RedisClient.srem(identifier, *values)

    @classmethod
    def clear(cls, identifier):
        cls.RedisClient.delete(cls.build_identifier(identifier))

    @classmethod
    def exists(cls, identifier, value):
        identifier = cls.build_identifier(identifier)
        value = cls.encrypt_data(identifier, value)
        return bool(cls.RedisClient.sismember(identifier, value))

    @classmethod
    def list_all(cls, identifier):
        return cls.RedisClient.smembers(cls.build_identifier(identifier))

    @classmethod
    def length(cls, identifier):
        return cls.RedisClient.scard(cls.build_identifier(identifier))



import datetime
import struct


def b2s(b):
    if isinstance(b, str):
        return b
    if not isinstance(b, bytes):
        raise TypeError("Invalid Bytes type")
    return b.decode("utf-8")


def require_string(a, int_to_string=False):
    if int_to_string and isinstance(a, int):
        a = str(a)
    if isinstance(a, bytes):
        a = b2s(a)
    if not isinstance(a, str):
        raise TypeError(f"A string object is required, not {str(type(a))}")
    return a


def b2timestamp(b):
    if not isinstance(b, bytes):
        raise TypeError("Invalid Bytes type")
    return struct.unpack(">Q", b)[0]


def timestamp2b(timestamp):
    if isinstance(timestamp, bytes):
        return timestamp
    if isinstance(timestamp, datetime.datetime):
        timestamp = int(timestamp.timestamp())
    if isinstance(timestamp, float):
        timestamp = int(timestamp)
    if not isinstance(timestamp, int):
        raise TypeError("Invalid Timestamp Type")
    return struct.pack(">Q", timestamp)


def require_timestamp_bytes(a):
    if isinstance(a, (int, datetime.datetime)):
        return timestamp2b(a)
    if not isinstance(a, bytes):
        raise TypeError(f"A byte object is required, not {str(type(a))}")
    return a


def s2b(s):
    if isinstance(s, bytes):
        return s
    if not isinstance(s, str):
        raise TypeError("Invalid String type")
    return s.encode("utf-8")


def require_bytes(a, int_to_string=False):
    if int_to_string and isinstance(a, int):
        a = str(a)
    if isinstance(a, str):
        return s2b(a)
    if not isinstance(a, bytes):
        raise TypeError(f"A byte object is required, not {str(type(a))}")
    return a



from .base_entity import BaseRedisEntity, RedisEntityMetaclass
from .parameters import require_string, require_bytes, b2timestamp, timestamp2b


class RedisHashmapEntity(BaseRedisEntity):
    """
    This class represents a HashMap that is stored in Redis.
    The HashMap is identified by a Prefix that specifies the name of the HashMap.
    A HashMap entry is identified by the Prefix and a unique identifier.
    The RedisHashmapEntity stores and loads all HashMap keys as strings, and all HashMap values as bytes.
    """

    Contents = {}

    @classmethod
    def _encrypt_kwargs_dict(cls, kwargs_dict):
        return {
            require_string(key): cls.encrypt_data(key, value)
            for key, value in kwargs_dict.items()
        }

    @classmethod
    def _decrypt_kwargs_dict(cls, kwargs_dict):
        return {
            require_string(key): cls.decrypt_data(key, value)
            for key, value in kwargs_dict.items()
        }

    @classmethod
    def store(cls, identifier, **kwargs):
        for keyword_arg in cls.Contents:
            if keyword_arg not in kwargs:
                raise AttributeError(
                    f"Key={keyword_arg} not specified. Please specify it through an keyword argument."
                )
        kwargs = cls._encrypt_kwargs_dict(kwargs)
        identifier = cls.build_identifier(identifier)
        cls.RedisClient.hset(identifier, mapping=kwargs)
        cls.add_expiration(identifier)
        return cls(identifier, **kwargs)

    @classmethod
    def exists(cls, identifier):
        return cls.RedisClient.exists(cls.build_identifier(identifier)) == 1

    @classmethod
    def load(cls, identifier):
        identifier = cls.build_identifier(identifier)
        return cls(
            identifier,
            **cls._decrypt_kwargs_dict(cls.RedisClient.hgetall(identifier)),
        )

    @classmethod
    def length(cls, identifier):
        return cls.RedisClient.hlen(cls.build_identifier(identifier))

    @classmethod
    def delete(cls, identifier):
        return cls.RedisClient.delete(cls.build_identifier(identifier))

    def __init__(self, identifier, **kwargs):
        self.identifier = identifier
        self.items = kwargs
        self._enforce_valid_entity()

    def __getattr__(self, item):
        if item in self.items:
            return self.items.get(item)
        raise AttributeError("Item not Found")

    def __delattr__(self, item):
        if item in self.items:
            self.items.pop(item)

    def _enforce_valid_entity(self):
        for keyword in self.Contents:
            if not hasattr(self, keyword):
                raise AttributeError(f"Key={keyword} not contained in Entity")



from .base_entity import BaseRedisEntity


class RedisListEntity(BaseRedisEntity):
    @classmethod
    def lpush(cls, identifier, *values):
        identifier = cls.build_identifier(identifier)
        values = [cls.encrypt_data(identifier, value) for value in values]
        cls.RedisClient.lpush(identifier, *values)
        cls.add_expiration(identifier)

    @classmethod
    def brpop(cls, identifier, timeout=0):
        identifier = cls.build_identifier(identifier)
        key, value = cls.RedisClient.brpop(identifier, timeout=timeout)
        return key, cls.decrypt_data(identifier, value)

    @classmethod
    def lindex(cls, identifier, index):
        identifier = cls.build_identifier(identifier)
        value = cls.RedisClient.lindex(identifier, index)
        return cls.decrypt_data(identifier, value)

    @classmethod
    def length(cls, identifier):
        return cls.RedisClient.llen(cls.build_identifier(identifier))

    @classmethod
    def clear(cls, identifier):
        return cls.RedisClient.delete(cls.build_identifier(identifier))



import hashlib
import json
import time
import base64
from secrets import compare_digest
from Crypto.Cipher import AES
from ..exceptions import DecryptionException
from .mixin_base import MixinBaseClass
from ..parameters import timestamp2b, b2timestamp, require_bytes
from ..exceptions import DecryptionException


class AuthenticatedEncryptionMixin(MixinBaseClass):
    AesKey = b""

    @classmethod
    def encrypt_data(cls, key, value) -> bytes:
        expire = cls.Expire if cls.Expire is not None else 0xFFFFFFFF
        expire = time.time() + expire
        aes = AES.new(cls.AesKey, AES.MODE_GCM, mac_len=16)
        aes.update(require_bytes(key))
        payload = timestamp2b(expire) + require_bytes(value)
        ciphertext, tag = aes.encrypt_and_digest(payload)
        nonce = aes.nonce
        payload = {
            "n": base64.b64encode(nonce).decode(),
            "c": base64.b64encode(ciphertext).decode(),
            "t": base64.b64encode(tag).decode(),
        }
        payload = json.dumps(payload).encode()
        return base64.b64encode(payload)

    @classmethod
    def decrypt_data(cls, key, value) -> bytes:
        payload = base64.b64decode(require_bytes(value))
        payload = json.loads(payload)
        for required_key in ("n", "c", "t"):
            if required_key not in payload:
                raise DecryptionException(f"Key {required_key} not in JSON payload")

        nonce = base64.b64decode(payload["n"])
        ciphertext = base64.b64decode(payload["c"])
        tag = base64.b64decode(payload["t"])

        try:
            aes = AES.new(cls.AesKey, AES.MODE_GCM, mac_len=16, nonce=nonce)
            aes.update(require_bytes(key))
            plaintext = aes.decrypt_and_verify(ciphertext, tag)
        except ValueError:
            raise DecryptionException("Decryption Failed")

        expire_timestamp = b2timestamp(plaintext[:8])
        dec_value = plaintext[8:]

        if time.time() > expire_timestamp:
            raise DecryptionException("Value is already marked as expired")

        return dec_value



from .hashed_identifier_mixin import HashedIdentifierMixin
from .authenticated_encryption_mixin import AuthenticatedEncryptionMixin
from .deterministic_authenticated_encryption_mixin import (
    DeterministicAuthenticatedEncryptionMixin,
)



#!/usr/bin/env python

from setuptools import setup
import pathlib


here = pathlib.Path(__file__).parent.resolve()
long_description = (here / "README.md").read_text(encoding="utf-8")

setup(
    name="redis_entities",
    version="1.0.5",
    description="Redis Entities",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Tibotix",
    author_email="tizian@seehaus.net",
    url="https://github.com/tibotix/redis_entities",
    package_dir={"redis_entities": "src"},
    packages=["redis_entities", "redis_entities.mixins"],
    extras_require={"test": ["pytest", "fakeredis"]},
    python_requires=">=3.5, <4",
)



import hashlib

from .mixin_base import MixinBaseClass
from ..parameters import require_bytes


class HashedIdentifierMixin(MixinBaseClass):
    HashName = "sha512_256"
    Salt = b""

    @classmethod
    def _hash_identifier(cls, identifier) -> str:
        return hashlib.new(
            cls.HashName, require_bytes(cls.Salt) + require_bytes(identifier)
        ).hexdigest()
